#!/usr/bin/env python
"""Processors used only by SAXSWAXS experiments."""

from typing import (
    Literal,
    Optional,
    Union,
)

from pydantic import (
    conint,
    conlist,
    Field,
    FilePath,
)
import numpy as np

from CHAP import Processor
from CHAP.common import ExpressionProcessor
from CHAP.common.models.map import (
    Detector,
    MapConfig,
)
from CHAP.common.models.integration import PyfaiIntegrationConfig


class CfProcessor(Processor):
    """Processor to calculate the correction factor Cf that, when
    multiplied by appropriately processed SAXSWAXS data obtained,
    converts data to absolute cross-section / intensity in inverse cm.
    """
    def process(
            self, data, interactive=False, save_figures=True, nxpath=None,
            radial_range=None, scan_step_indices=None, eps=1.e-5):
        """Return a dictionary with the computed correction factor Cf
        and the configuration parameters.

        :param data: Input data list containing the reference data
            labelled with `'reference_data'` as well as the NeXus
            input data with the azimuthally integrated SAXSWAXS data.
        :type data: list[PipelineData]
        :param interactive: Allows for user interactions,
            defaults to `False`.
        :type interactive: bool, optional
        :param save_figures: Create Matplotlib correction factor
            image that can be saved to file downstream in the workflow,
            defaults to `True`.
        :type save_figures: bool, optional
        :param nxpath: The path to a specific NeXus NXdata object in
            the input NeXus file tree to the measured data from.
        :type nxpath: str, optional
        :param radial_range: q-range used to compute Cf.
        :type radial_range: Union(
            list[float, float], tuple[float, float]), optional
        :ivar scan_step_indices: Optional scan step indices to use for
            the calculation. If not specified, the correction factor
            will be computed on the average of all data for the scan.
        :type scan_step_indices: int, str, list[int], optional
        :param eps: Minimum plotting value of the corrected azimuthally
            integrated SAXSWAXS data, default to `1.e-5`.
        :type eps: float
        :returns: Computed correction factor Cf and the configuration
            parameters plus the optional correction factor image as a
            CHAP.pipeline.PipelineData object.
        :rtype: Union[dict, (dict, PipelineData)]
        """
        # Third party modules
        from pandas import DataFrame
        from scipy.interpolate import interp1d

        if interactive or save_figures:
            # Third party modules
            import matplotlib.pyplot as plt

            # Local modules
            from CHAP.pipeline import PipelineData
            from CHAP.utils.general import (
                fig_to_iobuf,
                round_to_n,
            )

        # Validate the input parameters
        if scan_step_indices is not None:
            if isinstance(scan_step_indices, int):
                scan_step_indices = [scan_step_indices]
            elif isinstance(scan_step_indices, str):
                # Local modules
                from CHAP.utils.general import string_to_list

                scan_step_indices = string_to_list(scan_step_indices)

        # Load the measured data
        if nxpath is None:
            try:
                nxdata = self.get_default_nxdata(self.get_data(data))
            except Exception as exc:
                raise ValueError(
                    'No valid default NXdata object in pipeline data') from exc
        else:
            # Third party modules
            from nexusformat.nexus import NXdata

            try:
                nxdata = self.get_data(data)[nxpath]
                assert isinstance(nxdata, NXdata)
            except Exception as exc:
                raise ValueError('No valid default NXdata object in pipeline '
                                 'data') from exc
        if scan_step_indices is not None and nxdata.nxsignal.ndim != 2:
            self.logger.warning(
                'Input parameters scan_step_indices for map with a number of '
                'independent dimensions other than 1')
            scan_step_indices = None
        if nxdata.nxsignal.ndim == 1:
            data_meas = nxdata.nxsignal.nxdata
        elif nxdata.nxsignal.ndim == 2 and scan_step_indices is not None:
            data_meas = \
                nxdata.nxsignal.nxdata[scan_step_indices,:].mean(axis=0)
        else:
            data_meas = nxdata.nxsignal.nxdata.mean(
                axis=tuple(range(nxdata.nxsignal.ndim-1)))
        q_meas = nxdata[nxdata.attrs['axes'][1]].nxdata

        # Load the reference data
        ddata = self.get_data(data, name='reference_data')
        try:
            assert isinstance(ddata, DataFrame)
            assert len(ddata.columns) in (2, 3)
            q_ref = ddata.values[:,0]
            data_ref = ddata.values[:,1]
        except Exception as exc:
            raise ValueError(
                'Invalid reference data format {type(ddata)}') from exc

        # Interpolate measured reference data onto reference q values
        mask = np.where(
            q_ref <= radial_range[1],
            np.where(q_ref >= radial_range[0], 1, 0), 0).astype(bool)
        if not q_ref[mask].size:
            raise ValueError(
                f'No reference values within specified radial range')
        func = interp1d(q_meas, data_meas, kind='cubic')
        data_meas_intpol = func(q_ref[mask])
        if not data_meas_intpol.size:
            raise ValueError(
                f'No measured values within specified radial range')

        # Get the correction factor
        ratio = data_ref[mask]/data_meas_intpol
        cf = ratio.mean()
        cf_stv = ratio.std()
        self.logger.info(f'correction factor Cf: {round_to_n(cf, 6):e} \u00B1 '
                         f'{round_to_n(100*cf_stv/cf, 2)}%')

        # Assemble result
        result = {
            'cf': float(cf),
            'error': float(cf_stv),
            'radial_range': radial_range,
        }

        # Plot the results
        data_meas *= cf
        figures = None
        if interactive or save_figures:
            fig, ax = plt.subplots(figsize=(11, 8.5))
            ax.plot(q_ref, data_ref, label='APS Reference Data')
            ax.plot(q_meas, data_meas, label=f'Corrected FMB Data/C$_f$')
            for v in radial_range:
                plt.axvline(v, color='r', linestyle='--')
            ax.set_yscale('log')
            ax.set_title(f'Absolute Intensity Calculation', fontsize='xx-large')
            ax.set_xlabel(r'{q_A^-1}', fontsize='x-large')
            ax.set_ylabel('Normalized Intensity', fontsize='x-large')
            min_x = q_meas.min()
            max_x = q_meas.max()
            delta_x = 0.1*(q_meas.max() - q_meas.min())
            ax.set_xlim((min_x-delta_x, max_x+delta_x))
            ax.set_ylim(
                (10**np.floor(np.log10(data_meas[data_meas>eps].min())),
                 10**np.floor(1+np.log10(data_meas.max()))))
            ax.legend(fontsize='x-large', edgecolor='grey')
            plt.annotate(
                f'C$_f$ = {round_to_n(cf, 6):e} $\pm$ '
                f'{round_to_n(100*cf_stv/cf, 2)}%\n'
                f'1/C$_f$ = {round_to_n(1/cf, 6):e}',
                (.65, .9), xycoords = 'axes fraction', fontsize='x-large',
                bbox=dict(facecolor='white', edgecolor='grey', pad=10.0))
            if save_figures:
                fig.tight_layout(rect=(0, 0, 1, 0.95))
                figures  = fig_to_iobuf(fig)
            if interactive:
                plt.show()
            plt.close()

        if figures is not None:
            return (
                result,
                PipelineData(name=self.__name__, data=figures,
                    schema='common.write.ImageWriter'))
        return result


class FluxCorrectionProcessor(ExpressionProcessor):
    """Processor for flux correction."""
    def process(self, data, presample_intensity_reference_rate=None,
                nxprocess=False):
        """Given input data for `'intensity'`,
        `'presample_intensity'`, and `'dwell_time_actual'`, return
        flux corrected intensity signal.

        :param data: Input data list containing items with names
            `'intensity'`, `'presample_intensity'`, and (if
            `presample_intensity_reference_rate` is not specified)
            `'dwell_time_actual'`.
        :type data: list[PipelineData]
        :param presample_intensity_reference_rate: Reference counting
            rate for the `'presample_intensity'` signal. If not
            specified, it will be calculated with
            `'np.nanmean(presample_intensity /
            dwell_time_actual)'`. Defaults to `None`.
        :type presample_intensity_reference_rate: float, optional
        :param nxprocess: Flag to indicate the flux corrected data
            should be returned as an `NXprocess`. Defaults to `False`.
        :returns: Flux corrected version of input `'intensity'` data.
        :rtype: object
        """
        if presample_intensity_reference_rate is None:
            presample_intensity_reference_rate = self._process(
                data,
                'np.nanmean(presample_intensity / dwell_time_actual)'
            )
        presample_intensity = self.get_data(
            data, name='presample_intensity',
        )
        intensity = self.get_data(
            data, name='intensity',
        )
        # nxfieldtable = {
        #     'intensity': intensity,
        #     'presample_intensity': presample_intensity,
        #     'presample_intensity_reference_rate': NXfield(
        #         name='presample_intensity_reference_rate',
        #         values=presample_intensity_reference_rate
        #     )
        # }

        # Extend presample_intensity along last dim to have same shape
        # as intensity
        for dim in intensity.shape[presample_intensity.ndim:]:
            presample_intensity = np.repeat(
                np.expand_dims(presample_intensity, axis=-1), dim, axis=-1
            )
        symtable = {
            'presample_intensity_reference_rate':
                presample_intensity_reference_rate,
            'intensity': intensity,
            'presample_intensity': presample_intensity
        }
        expression = (
            'intensity *'
            '(presample_intensity_reference_rate / presample_intensity)'
        )
        return self._process(
            data, expression, symtable=symtable, nxprocess=nxprocess,
            nxfieldtable={}
        )


class FluxAbsorptionCorrectionProcessor(ExpressionProcessor):
    """Processor for flux and absorption correction."""
    def process(self, data,
                presample_intensity_reference_rate=None,
                nxprocess=False):
        """Given input data for `'intensity'`,
        `'presample_intensity'`, `'postsample_intensity'`,
        `'background_presample_intensity'`,
        `'background_postsample_intensity'`, and
        `'dwell_time_actual'`, return flux and absorption corrected
        intensity signal.

        :param data: Input data list containing all necessary data
            labelled with their proper names.
        :type data: list[PipelineData]
        :param presample_intensity_reference_rate: Reference counting
            rate for the `'presample_intensity'` signal. If not
            specified, it will be calculated with
            `'np.nanmean(presample_intensity /
            dwell_time_actual)'`. Defaults to `None`.
        :type presample_intensity_reference_rate: float, optional
        :param nxprocess: Flag to indicate the flux corrected data
            should be returned as an `NXprocess`. Defaults to `False`.
        :returns: Flux and absprption corrected version of input
            `'intensity'` data.
        :rtype: object
        """
        intensity = self.get_data(
            data, name='intensity',
        )

        if presample_intensity_reference_rate is None:
            presample_intensity_reference_rate = self._process(
                data,
                'np.nanmean(presample_intensity / dwell_time_actual)'
            )

        T = self._process(
            data,
            ('(postsample_intensity / presample_intensity) '
             '/ np.average('
             '(background_postsample_intensity / background_presample_intensity))')
        )
        # Extend T along last dim to have same shape as intensity
        for dim in intensity.shape[T.ndim:]:
            T = np.repeat(
                np.expand_dims(T, axis=-1), dim, axis=-1
            )

        presample_intensity = self.get_data(
            data, name='presample_intensity',
        )
        # Extend presample_intensity along last dim to have same shape
        # as intensity
        for dim in intensity.shape[presample_intensity.ndim:]:
            presample_intensity = np.repeat(
                np.expand_dims(presample_intensity, axis=-1), dim, axis=-1
            )

        symtable = {
            'presample_intensity_reference_rate':
                presample_intensity_reference_rate,
            'intensity': intensity,
            'presample_intensity': presample_intensity,
            'T': T
        }
        expression = (
            '(1 / T)'
            '* intensity'
            '* (presample_intensity_reference_rate / presample_intensity)'
        )
        return self._process(
            data, expression, symtable=symtable, nxprocess=nxprocess,
            nxfieldtable={}
        )


class FluxAbsorptionBackgroundCorrectionProcessor(ExpressionProcessor):
    """Processor for flux, absorption, and background correction. May
    also perform thickness correction."""
    def process(self, data,
                presample_intensity_reference_rate=None,
                sample_thickness_cm=None,
                sample_mu_inv_cm=None,
                nxprocess=False):
        """Given input data for `'intensity'`,
        `'presample_intensity'`, `'postsample_intensity'`,
        `'background_presample_intensity'`,
        `'background_postsample_intensity'`, `'background_intensity'`,
        and `'dwell_time_actual'`, return flux and absorption
        corrected intensity signal.

        :param data: Input data list containing all necessary data
            labelled with their proper names.
        :type data: list[PipelineData]
        :param presample_intensity_reference_rate: Reference counting
            rate for the `'presample_intensity'` signal. If not
            specified, it will be calculated with
            `'np.nanmean(presample_intensity /
            dwell_time_actual)'`. Defaults to `None`.
        :type presample_intensity_reference_rate: float, optional
        :param sample_thickness_cm: Sample thickness in
            centimeters. If specified, this processor will
            additionally perform thickness correction. Use of this
            parameter is mutualy exclusive with
            use of `sample_mu_inv_cm`. Defaults to `None`.
        :type sample_thickness_cm: float, optional
        :param sample_mu_inv_cm: Sample linear attenuation coefficient
            in inverse centimeters. If specified, this processor will
            additionally perform thickness correction. Use of this
            parameter is mutualy exclusive with use of
            `sample_thickness_cm`. Defaults to `None`.
        :type sample_mu_inv_cm: float, optional
        :param nxprocess: Flag to indicate the flux corrected data
            should be returned as an `NXprocess`. Defaults to `False`.
        :returns: Flux and absprption corrected version of input
            `'intensity'` data.
        :rtype: object
        """
        if sample_thickness_cm is not None and sample_mu_inv_cm is not None:
            raise ValueError((
                'Cannot use sample_thickness_cm and sample_mu_inv_cm'
                ' at the same time'
            ))

        intensity = self.get_data(
            data, name='intensity',
        )

        if presample_intensity_reference_rate is None:
            presample_intensity_reference_rate = self._process(
                data,
                'np.nanmean(presample_intensity / dwell_time_actual)'
            )

        T = self._process(
            data,
            ('(postsample_intensity / presample_intensity) '
             '/ np.average('
             '(background_postsample_intensity / background_presample_intensity))')
        )
        # Extend T along last dim to have same shape as intensity
        for dim in intensity.shape[T.ndim:]:
            T = np.repeat(
                np.expand_dims(T, axis=-1), dim, axis=-1
            )

        presample_intensity = self.get_data(
            data, name='presample_intensity',
        )
        # Extend presample_intensity along last dim to have same shape
        # as intensity
        for dim in intensity.shape[presample_intensity.ndim:]:
            presample_intensity = np.repeat(
                np.expand_dims(presample_intensity, axis=-1), dim, axis=-1
            )

        # Broadcast background intensity signal to shape of measured
        # intensity signal
        background_intensity = self.get_data(
            data, name='background_intensity',
        )
        background_intensity = np.broadcast_to(
            background_intensity, intensity.shape)

        if sample_thickness_cm is not None:
            t = sample_thickness_cm
        elif sample_mu_inv_cm is not None:
            t = -np.log(T / sample_mu_inv_cm)
        else:
            t = 1

        symtable = {
            't': t,
            'presample_intensity_reference_rate':
                presample_intensity_reference_rate,
            'intensity': intensity,
            'presample_intensity': presample_intensity,
            'T': T,
            'background_intensity': background_intensity
        }
        expression = (
            '(1 / t)'
            '* ('
            '(1 / T)'
            '* intensity'
            '* (presample_intensity_reference_rate / presample_intensity)'
            ') - ('
            'background_intensity'
            '* (presample_intensity_reference_rate / np.average(background_presample_intensity))'
            ')'
        )
        return self._process(
            data, expression, symtable=symtable, nxprocess=nxprocess,
            nxfieldtable={}
        )


class PyfaiIntegrationProcessor(Processor):
    """Processor for performing pyFAI integrations.

    :ivar config: PyfaiIntegrationConfig
    :type config: CHAP.common.models.integration.PyfaiIntegrationConfig
    """
    pipeline_fields: dict = Field(
        default={
            'config': 'common.models.integration.PyfaiIntegrationConfig'
        },
        init_var=True)
    config: PyfaiIntegrationConfig

    def process(self, data,
                idx_slices=[{'start':0, 'step': 1}]):
        """Perform a set of integrations on 2D detector data.

        :param data: input 2D detector data
        :type data: list[PipelineData]
        :param idx_slices: List of dicionaries identifying the sliced
            index at which the output data should be written in a
            dataset. Optional.
        :type idx_slices: list[dict[str, int]], defaults to
        `[{'start':0, 'step': 1}]`
        :return: List of dictionaries ready for use with
            `saxswaxs.ZarrResultsWriter` or
            `saxswaxs.NexusResultsWriter`.
        :rtype: list[dict[str, object]]
        """
        import time

        # Organize input for integrations
        input_data = {d['name']: d['data']
            for d in [d for d in data if isinstance(d['data'], np.ndarray)]}
        ais = {ai.get_id(): ai for ai in self.config.azimuthal_integrators}

        # Finalize idx slice for results
        idx = tuple(slice(idx_slice.get('start'),
                     idx_slice.get('stop'),
                     idx_slice.get('step')) for idx_slice in idx_slices)

        # Perform integration(s), package results for ZarrResultsWriter
        results = []
        nframes = len(input_data[list(input_data.keys())[0]])
        for integration in self.config.integrations:
            t0 = time.time()
            self.logger.info(f'Integrating {integration.name}...')
            result = integration.integrate(ais, input_data)
            tf = time.time()
            self.logger.debug(
                f'Integrated {integration.name} '
                f'({nframes/(tf-t0):.3f} frames/sec)')
            results.extend(
                [
                    {
                        'path': f'{integration.name}/data/I',
                        'idx': idx,
                        'data': np.asarray(result['intensities']),
                    },
                ]
            )
        return results


class SetupResultsProcessor(Processor):
    """Processor for creating an intital zarr structure with empty datasets
    for filling in by `saxswaxs.PyfaiIntegrationProcessor` and
    `common.ZarrValuesWriter`.

    :ivar dataset_shape: Shape of the completed dataset that will
        be processed later on (shape of the measurement itself,
        _not_ including the dimensions of any signals collected at
        each point in that measurement).
    :type dataset_shape: Union[int, list[int]]
    :ivar dataset_chunks: Extent of chunks along each dimension
        of the completed dataset / measurement. Choose this
        according to how you will process your data -- for
        example, if your `dataset_shape` is `[m, n]`, and you are
        planning to process each of the `m` rows as chunks,
        `dataset_chunks` should be `[1, n]`. But if you plan to
        process each of the `n` columns as chunks,
        `dataset_chunks` should be `[m, 1]`.
    :type dataset_chunks: Union[list[int], Literal["auto"]]
    """
    pipeline_fields: dict = Field(
        default={
            'config': 'common.models.integration.PyfaiIntegrationConfig'
        },
        init_var=True)
    config: PyfaiIntegrationConfig
    dataset_shape: conlist(item_type=conint(gt=0), min_length=1)
    dataset_chunks: Optional[
        Union[
            Literal['auto'],
            conlist(item_type=conint(gt=0), min_length=1)
        ]] = 'auto'

    def process(self, data):
        """Return a `zarr.group` to hold processed SAXS/WAXS data
        processed by `saxswaxs.PyfaiIntegrationProcessor`.

        :param data: Input data (configurations).
        :type data: list[PipelineData]
        :return: Empty structure for filling in SAXS/WAXS data
        :rtype: zarr.group
        """

        # Get zarr tree as dict from the
        # PyfaiIntegrationConfig
        tree = self.config.zarr_tree(self.dataset_shape, self.dataset_chunks)

        # Construct & return the root zarr.group
        return self.zarr_setup(tree)

    def zarr_setup(self, tree):
        """Return a `zarr.group` based on a
        dictionary representing a zarr tree of groups and arrays.

        :param tree: Nested dictionary representing a zarr tree of
            groups and arrays.
        :type tree: dict[str, object]
        :return: Zarr group corresponding to the contents of `tree`.
        :rtype: zarr.group
        """
        import zarr
        from zarr.storage import MemoryStore

        def create_group_or_dataset(node, zarr_parent, indent=0):
            # Set attributes if present
            if 'attributes' in node:
                for key, value in node['attributes'].items():
                    zarr_parent.attrs[key] = value
            # Create children (groups or datasets)
            if 'children' in node:
                for name, child in node['children'].items():
                    if 'shape' in child or 'data' in child:
                        # It's a dataset
                        self.logger.debug(f'Adding dset: {name}')
                        zarr_parent.create_dataset(
                            name,
                            **child,
                        )
                        # Set dataset attributes
                        if 'attributes' in child:
                            for key, value in child['attributes'].items():
                                zarr_parent[name].attrs[key] = value
                    else:
                        # It's a group
                        group = zarr_parent.create_group(name)
                        create_group_or_dataset(child, group, indent=indent+2)
        results = zarr.create_group(store=MemoryStore({}))
        create_group_or_dataset(tree['root'], results)
        return results


class SetupProcessor(Processor):
    """Convenience Processor for setting up a container for SAXS/WAXS
    experiments.

    :ivar detectors: List of basic detector configuration parameters.
    :type detectors: `CHAP.common.models.map.DetectorConfig`
    :ivar dataset_shape: Shape of the completed dataset that will
        be processed later on (shape of the measurement itself,
        _not_ including the dimensions of any signals collected at
        each point in that measurement).
    :type dataset_shape: Union[int, list[int]]
    :ivar dataset_chunks: Extent of chunks along each dimension
        of the completed dataset / measurement. Choose this
        according to how you will process your data -- for
        example, if your `dataset_shape` is `[m, n]`, and you are
        planning to process each of the `m` rows as chunks,
        `dataset_chunks` should be `[1, n]`. But if you plan to
        process each of the `n` columns as chunks,
        `dataset_chunks` should be `[m, 1]`.
    :type dataset_chunks: Union[list[int], Literal["auto"]]
    :ivar num_chunk: Used only if `dataset_chunks` is
        `"auto"`. Preferred number of chunks in the dataset. Defaults
        to `1`.
    :type num_chunk: int, optional
    :ivar raw_data: Flag to indicate wether or not space for raw
        detector data should be included in the returned Zarr
        structure; defaults to `True`.
    :type raw_data: bool, optional
    """
    pipeline_fields: dict = Field(
        default={
            'map_config': 'common.models.map.MapConfig',
            'pyfai_config': 'common.models.integration.PyfaiIntegrationConfig'
        },
        init_var=True)
    # map_config needs a default value because the map configuration
    # may be created by an earlier item in the Pipeline. If this is
    # the case, then map_config needs SOME value in order for the
    # Pipeline to pass validation.
    map_config: MapConfig = None
    pyfai_config: PyfaiIntegrationConfig
    detectors: conlist(item_type=Detector, min_length=1)
    dataset_shape: Optional[
        conlist(item_type=conint(gt=0), min_length=1)] = None
    dataset_chunks: Optional[
        Union[
            Literal['auto'],
            conlist(item_type=conint(gt=0), min_length=1)
        ]] = 'auto'
    num_chunk: Optional[conint(gt=0)] = 1
    raw_data: Optional[bool] = True

    def process(self, data):
        import asyncio
        import logging
        import zarr
        from zarr.core.buffer import default_buffer_prototype
        from zarr.storage import MemoryStore

        from CHAP.pipeline import PipelineData
        from CHAP.common import MapProcessor, NexusToZarrProcessor
        from CHAP.saxswaxs import SetupResultsProcessor

        def set_logger(pipeline_item):
            pipeline_item.logger = self.logger
            pipeline_item.logger.name = pipeline_item.__class__.__name__
            handler = logging.StreamHandler()
            handler.setFormatter(logging.Formatter(
                '{asctime}: {name:20} (from '+ self.__class__.__name__
                + '): {levelname}: {message}',
                datefmt='%Y-%m-%d %H:%M:%S', style='{'))
            pipeline_item.logger.removeHandler(
                pipeline_item.logger.handlers[0])
            pipeline_item.logger.addHandler(handler)
            return pipeline_item

        # Get NXroot container for raw data map
        map_processor_kwargs = {
            'config': self.map_config
        }
        if self.raw_data:
            map_processor_kwargs['detector_config'] = {
                'detectors': self.detectors
            }
        else:
            map_processor_kwargs['detector_config'] = {
                'detectors': []
            }
        setup_map_processor = set_logger(
            MapProcessor(
                **map_processor_kwargs
                # config=self.map_config,
                # detector_config={'detectors': self.detectors},
            )
        )
        ddata = [
            PipelineData(
                data=setup_map_processor.process(
                    data=data, fill_data=False),
            )
        ]

        if self.dataset_shape is None:
            try:
                nxroot = self.get_data(ddata, remove=False)
                nxentry = self.get_default_nxentry(nxroot)
                nxdata = nxentry[nxentry.default]
                for detector in self.detectors:
                    if self.dataset_shape is None:
                        self.dataset_shape = nxdata[
                            detector.get_id()].shape[:-2]
                    else:
                        assert (
                            self.dataset_shape == nxdata[
                                detector.get_id()].shape[:-2]
                        )
            except Exception as exc:
                raise ValueError(
                    'Unable to get consistent dataset shape from map') from exc
        if self.dataset_chunks == 'auto':
            self.dataset_chunks = self.dataset_shape[0]//self.num_chunk
            if self.num_chunk*self.dataset_chunks < self.dataset_shape[0]:
                self.dataset_chunks += 1
            self.dataset_chunks = [self.dataset_chunks]

        # Convert raw data map container to zarr format
        ddata_converter = set_logger(NexusToZarrProcessor())
        zarr_map = ddata_converter.process(ddata, chunks=self.dataset_chunks)

        # Get zarr container for integration results
        setup_results_processor = set_logger(
            SetupResultsProcessor(
                config=self.pyfai_config,
                dataset_shape=self.dataset_shape,
                dataset_chunks=self.dataset_chunks,
            )
        )
        zarr_results = setup_results_processor.process(data)

        # Assemble containers for raw & processed data
        zarr_root = zarr.create_group(store=MemoryStore({}))
        async def copy_zarr_store(source_store, dest_store):
            async for k in source_store.list():
                self.logger.info(f'Copying {k}')
                buf = await source_store.get(
                    k, prototype=default_buffer_prototype())
                await dest_store.set(k, buf)
        asyncio.run(copy_zarr_store(zarr_map.store, zarr_root.store))
        asyncio.run(copy_zarr_store(zarr_results.store, zarr_root.store))
        return zarr_root


class UnstructuredToStructuredProcessor(Processor):
    """Processor to aggregate "unstructured" data into a single NXdata
    with a "structured" representation.
    """
    def process(self, data, fields, name='data', attrs=None):
        """Return an `NXdata` object containing a single structured
        dataset composed from multiple unstructured input datasets.

        This method validates the field configuration, validates and
        reshapes the input data, determines common axes across all
        signals, and constructs a NeXus `NXdata` group containing
        signal and axis fields.

        :param data: Input data objects containing unstructured
            datasets.
        :type data: list[PipelineData]
        :param fields: Configuration describing how to structure the
            input data.  This is a list of dictionaries. Each
            dictionary must contain the required keys:

            - ``"name"``: Name of the data item, which must correspond
              to the ``name`` field of an item in ``data``.
            - ``"type"``: Either ``"signal"`` or ``"axis"``.
            - ``"axes"``: Required only for items where ``"type"`` is
              ``"signal"``. List of the names of the fields containing
              coordinate axes data for each dimension of the signal.

            Optional keys include:
            - ``"attrs"``: Dictionary of NeXus attributes to attach to

        :type fields: list[dict[str, object]]
        :param name: Name of the resulting `NXdata` group.
        :type name: str
        :param attrs: Attributes to attach to the resulting `NXdata`
            group.  The common axes determined during processing will
            be added to this dictionary under the ``"axes"`` key.
        :type attrs: dict[str, object] or None
        :returns: A structured NeXus `NXdata` object containing all
            signals and axes defined by the configuration.
        :rtype: nexusformat.nexus.NXdata
        """
        from nexusformat.nexus import NXdata, NXfield


        signals, axes = self.validate_config_fields(fields)
        signals, axes = self.validate_data(data, signals, axes)
        common_axes = self.get_common_axes(signals)

        signals, axes, common_axes = self.structure_signal_values(
            signals, axes, common_axes)
        if attrs is None:
            attrs = {}
        attrs.update({'axes': common_axes})

        return NXdata(
            name=name,
            attrs=attrs,
            **{
                signal['name']: NXfield(
                    name=signal['name'],
                    attrs=signal['attrs'],
                    value=signal['value_structured']
                )
                for signal in signals
            },
            **{
                axis['name']: NXfield(
                    name=axis['name'],
                    attrs=axis['attrs'],
                    value=axis['value_unique']
                )
                for axis in axes
            }
        )

    def validate_config_fields(self, fields):
        """Validate and normalize the field configuration.

        This method separates the input field configuration into signal and
        axis definitions, performs basic validation, and ensures that all
        axes referenced by signals are defined as axis fields.

        The returned signal and axis dictionaries are normalized into a
        consistent internal representation used by later processing stages.

        :param fields: Configuration describing how input data should be
            structured. Each item must define a ``"name"`` and ``"type"``
            key, where ``"type"`` is either ``"signal"`` or ``"axis"``.
            Signal entries must additionally define an ``"axes"`` list.
        :type fields: list[dict[str, object]]
        :returns: Tuple of validated signal and axis definitions.
        :rtype: tuple[list[dict], list[dict]]
        :raises ValueError: If a signal references an axis that is not defined,
            or if a signal is defined before any axes exist.
        """
        self.logger.info('Validating fields parameter')

        axes = []
        signals = []

        for field in fields:
            field_type = field.get('type')
            name = field.get('name')
            attrs = field.get('attrs', {})

            if field_type == 'axis':
                axes.append({'name': name, 'value': None, 'attrs': attrs})
                self.logger.debug(f'Registered axis "{name}"')

            elif field_type == 'signal':
                _axes = field.get('axes', [])
                if not axes:
                    raise ValueError(f'Signal "{name}" has no axes defined')
                signals.append({'name': name, 'axes': _axes,
                                'value': None, 'attrs': attrs})
                self.logger.debug(
                    f'Registered signal "{name}" with axes {_axes}'
                )

        # Validate that all axes used by signals exist as type: axis
        axes_names = [a['name'] for a in axes]
        for signal in signals:
            for axis in signal['axes']:
                if axis not in axes_names:
                    raise ValueError(
                        f'Signal {signal["name"]} '
                        + f'references unknown axis "{axis}"'
                    )

        self.logger.info(
            'Validated configuration for '
            + f'{len(signals)} signals and {len(axes)} axes'
        )
        return signals, axes

    def get_common_axes(self, signals):
        """Determine the common leading axes shared by all signals.

        This method computes the longest common *prefix* of axis names
        across all signal definitions. Only axes that appear in the
        same order at the beginning of each signal's ``axes`` list are
        included in the result.

        This is used to identify the shared coordinate dimensions
        for a structured `NXdata` group.

        :param signals: Validated signal definitions. Each signal must
            define an ``"axes"`` key containing an ordered list of
            axis names.
        :type signals: list[dict]

        :returns: List of axis names that form the common leading axes
            for all signals. Returns an empty list if no common prefix
            exists.
        :rtype: list[str]
        """
        self.logger.info('Computing common dataset axes')

        if not signals:
            self.logger.warning('No signals provided; no common axes')
            return []

        # Start with the first signal's axes
        common_axes = list(signals[0]['axes'])

        for signal in signals[1:]:
            _axes = signal['axes']
            i = 0
            max_i = min(len(common_axes), len(_axes))
            while i < max_i and common_axes[i] == _axes[i]:
                i += 1
            common_axes = common_axes[:i]
            if not common_axes:
                break

        self.logger.info(f'Computed common axes: {common_axes}')
        return common_axes

    def validate_data(self, data, signals, axes):
        """Validate and normalize input data for axes and signals.

        This method retrieves raw input data for each axis and signal,
        propagates metadata attributes, computes unique axis values,
        and allocates structured arrays for signal data.

        For each axis:
          - The raw data is loaded
          - Attributes are merged (without overwriting user-specified
            ones)
          - Unique axis values are computed

        For each signal:
          - The raw data is loaded
          - Attributes are merged (without overwriting user-specified
            ones)
          - A structured output array is allocated based on its axes
          - The total signal size is validated against the expected
            shape

        :param data: Input unstructured data items.
        :type data: list[PipelineData]
        :param signals: Validated signal field definitions.
        :type signals: list[dict]
        :param axes: Validated axis field definitions.
        :type axes: list[dict]
        :returns: Updated signal and axis definitions with populated
            values and derived metadata.
        :rtype: tuple[list[dict], list[dict]]
        :raises ValueError: If a signal's data size does not match the
            expected size derived from its axes.
        """
        import numpy as np

        self.logger.info('Validating input data')
        self.logger.info('Validating axis data')
        for axis in axes:
            value = self.get_data(data, name=axis['name'])
            # Merge attributes, preserving explicitly defined ones
            axis['attrs'] = {
                **axis['attrs'],
                **{k: v for k, v in value.attrs.items()
                   if k not in axis['attrs'] and k != 'target'}
            }
            axis['value'] = value
            axis['value_unique'] = np.unique(value)
            self.logger.debug(
                f'Axis {axis["name"]}: {value.size} entries, '
                f'{axis["value_unique"].size} unique'
            )

        # Build a lookup table for faster axis access by name
        axes_by_name = {a['name']: a for a in axes}
        self.logger.info("Validating signal data")
        for signal in signals:
            name = signal['name']
            value = self.get_data(data, name=name)
            # Merge attributes, preserving explicitly defined ones
            signal['attrs'] = {
                **signal['attrs'],
                **{k: v for k, v in value.attrs.items()
                   if k not in signal['attrs'] and k != 'target'}
            }
            signal['value'] = value
            _axes = signal['axes']
            signal['attrs']['axes'] = _axes
            shape = tuple(
                [axes_by_name[a]['value_unique'].size for a in _axes]
            )
            signal['value_structured'] = np.empty(shape, dtype=value.dtype)
            size_expected = np.prod(shape)
            size_actual = signal['value'].size
            self.logger.debug(
                f'Signal "{name}": expected size {size_expected} (shape: {shape}), '
                f'actual size {size_actual} (shape: {value.shape})'
            )
            if size_actual != size_expected:
                raise(ValueError(
                    f'Signal {name} has size {size_actual}; '
                    + f'expected {size_expected}'
                ))
        self.logger.info('Validated input data')
        return signals, axes

    def structure_signal_values(self, signals, axes, common_axes):
        """Reshape and populate structured signal arrays using common axes.

        This method determines computes index mappings from raw axis
        values to their unique sorted representations, and inserts
        each signal's unstructured data into its preallocated
        structured array.

        Only the common axes are used for structuring; any trailing,
        signal-specific axes are assumed to have already been handled when
        allocating the structured signal arrays.

        :param signals: Signal definitions with raw and preallocated
            structured data arrays.
        :type signals: list[dict]
        :param axes: Axis definitions containing raw values and unique
            values.
        :type axes: list[dict]
        :param common_axes: Ordered list of the names of the dataset's
            common axes.
        :type common_axes: list[str]
        :returns:
            - Updated signal definitions with populated structured arrays
            - Unmodified axis definitions
            - List of common axis names shared by all signals
        :rtype: tuple[list[dict], list[dict], list[str]]
        """
        self.logger.info('Structuring dataset')
        axes_by_name = {a['name']: a for a in axes if a['name'] in common_axes}
        indices = {
            a: np.searchsorted(axes_by_name[a]['value_unique'],
                               axes_by_name[a]['value'])
            for a in common_axes
        }
        _indices = tuple(indices[a] for a in common_axes)
        for signal in signals:
            signal['value_structured'][_indices] = signal['value']

        return signals, axes, common_axes


class UpdateValuesProcessor(Processor):
    """Processes a slice of data for updating values in an existing
    container for a SAXS/WAXS experiment.

    :ivar map_config: Map Configuration.
    :type map_config: CHAP.common.models.map.MapConfig
    :ivar pyfai_config: PyFAI integration configuration.
    :type pyfai_config: CHAP.common.models.integration.PyfaiIntegrationConfig
    :ivar spec_file: SPEC file containing scan from which to read and
        process a slice of raw data.
    :type spec_file: str
    :ivar scan_number: Number of scan from which to read and process a
        slice of raw data.
    :type scan_number: int
    :ivar detectors: List of detector configurations.
    :type detectors: list[Detector]
    :ivar raw_data: Flag to indicate wether or not space for raw
        detector data should be included in the values returned;
        defaults to `True`.
    :type raw_data: bool, optional
    """
    pipeline_fields: dict = Field(
        default={
            'map_config': 'common.models.map.MapConfig',
            'pyfai_config': 'common.models.integration.PyfaiIntegrationConfig'
        },
        init_var=True)
    # map_config needs a default value because the map configuration
    # may be created by an earlier item in the Pipeline. If this is
    # the case, then map_config needs SOME value in order for the
    # Pipeline to pass validation.
    map_config: MapConfig = None
    pyfai_config: PyfaiIntegrationConfig
    spec_file: FilePath
    scan_number: conint(gt=0)
    detectors: conlist(item_type=Detector, min_length=1)
    raw_data: Optional[bool] = True

    def process(self, data, idx_slice={'start': 0, 'step': 1}):
        # Get updates with MapSliceProcessor
        # Pass detector data to PyfaiIntegration processor
        # Concatenate & return results
        import logging
        import os

        from CHAP.common import MapSliceProcessor
        from CHAP.pipeline import PipelineData
        from CHAP.saxswaxs import PyfaiIntegrationProcessor

        def set_logger(pipeline_item):
            pipeline_item.logger = self.logger
            pipeline_item.logger.name = pipeline_item.__class__.__name__
            handler = logging.StreamHandler()
            handler.setFormatter(logging.Formatter(
                '{asctime}: {name:20} (from '+ self.__class__.__name__
                + '): {levelname}: {message}',
                datefmt='%Y-%m-%d %H:%M:%S', style='{'))
            pipeline_item.logger.removeHandler(
                pipeline_item.logger.handlers[0])
            pipeline_item.logger.addHandler(handler)
            return pipeline_item

        # Read in slice of raw data
        raw_values = set_logger(
            MapSliceProcessor(
                map_config=self.map_config,
                detectors=self.detectors,
                spec_file=str(self.spec_file),
                scan_number=self.scan_number,
            )
        ).process(None, idx_slice=idx_slice)

        def get_detector_data(values, name):
            for v in values:
                if os.path.basename(v['path']) == name:
                    return v['data']
            return None

        # Use raw detector data as input to integration
        for d in self.detectors:
            data.append(
                PipelineData(
                    name=d.get_id(),
                    data=get_detector_data(raw_values, d.get_id()),
                )
            )
        # Get integrated data
        processed_values = set_logger(
            PyfaiIntegrationProcessor(
                config=self.pyfai_config,
            )
        ).process(data, idx_slices=[idx_slice])

        if self.raw_data:
            return raw_values + processed_values
        return processed_values


if __name__ == '__main__':
    # Local modules
    from CHAP.processor import main

    main()



#!/usr/bin/env python
#-*- coding: utf-8 -*-
"""
File       : processor.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: Processor module for FOXDEN services
"""

# Local modules
from CHAP.common.utils import (
    osinfo,
    environments,
)
from CHAP.processor import Processor


#class FoxdenMetadataProcessor(Processor):
#    """Processor to collect CHAP workflow metadata from a workflow
#    NeXus output object.
#    """
#    def process(self, data):
#        """Extract metadata from a workflow NeXus output object for
#        submission to the FOXDEN Metadata service.
#
#        :param data: Input data.
#        :type data: list[PipelineData]
#        :return: CHAP workflow metadata record.
#        :rtype: dict
#        """
#        # Third party modules
#        from json import loads
#        from nexusformat.nexus import (
#            NXentry,
#            NXroot,
#        )
#
#        # Load and validate the workflow NeXus output object
#        nxentry = self.get_data(data, remove=False)
#        if isinstance(nxentry, NXroot):
#            nxentry = nxentry[nxentry.default]
#        if not isinstance(nxentry, NXentry):
#            raise ValueError(f'Invalid input data type {type(nxentry)}')
#
#        # Get did and experiment type
#        map_config = loads(str(nxentry.map_config))
#        did = map_config['did']
#        experiment_type = map_config['experiment_type']
#
#        # Extract metadata
#        method = getattr(self, f'_get_metadata_{experiment_type.lower()}')
#        metadata = method(nxentry)
#
#        if 'reconstructed_data' in metadata:
#            did = f'{did}/{experiment_type.lower()}_reconstructed'
#        else:
#            did = f'{did}/{experiment_type.lower()}_reduced'
#        return {'did': did, 'application': 'CHAP', 'metadata': metadata}
#
#    def _get_metadata_tomo(self, nxentry):
#        metadata = {}
#        if 'reduced_data' in nxentry:
#            data = nxentry.reduced_data
#            metadata.update({
#                'reduced_data': {
#                    'date': str(data.date),
#                    'img_row_bounds': data.img_row_bounds.tolist(),
#                }
#            })
#        if 'reconstructed_data' in nxentry:
#            data = nxentry.reconstructed_data
#            metadata.update({
#                'reconstructed_data': {
#                    'date': str(data.date),
#                    'center_offsets': data.center_offsets.tolist(),
#                    'center_rows': data.center_offsets.tolist(),
#                    'center_stack_index': int(data.center_stack_index),
#                    'x_bounds': data.x_bounds.tolist(),
#                    'y_bounds': data.y_bounds.tolist(),
#                }
#            })
#        if 'combined_data' in nxentry:
#            data = nxentry.combined_data
#            metadata.update({
#                'combined_data': {
#                    'date': str(data.date),
#                }
#            })
#        return metadata


class FoxdenProvenanceProcessor(Processor):
    """Processor to collect CHAP workflow provenance data."""
    def process(self, data):
        """Extract provenance data from the pipeline data for
        submission to the FOXDEN Provenance service.

        :param data: Input data.
        :type data: list[PipelineData]
        :return: CHAP workflow provenance record.
        :rtype: dict
        """
        # Load the provenance info
        provenance = self.get_data(data, schema='provenance')

        # Add system info to provenance data
        did = provenance['did']
        provenance.update({
            'parent_did': did.rsplit('/', 1)[0],
            'scripts': [
                {'name': 'CHAP', 'parent_script': None, 'order_idx': 1}],
            'site': 'Cornell',
            'osinfo': osinfo(),
            'environments': environments(),
            'processing': 'CHAP pipeline',
        })

        return provenance


if __name__ == '__main__':
    # Local modules
    from CHAP.processor import main

    main()



#!/usr/bin/env python
"""
File       : writer.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: generic Writer module

Define a generic `Writer` object.
"""

# System modules
import argparse
import logging
import os
from sys import modules
from typing import Optional

# Third party modules
from pydantic import model_validator

# Local modules
from CHAP.pipeline import PipelineItem


def validate_writer_model(writer):
    writer.filename = os.path.normpath(os.path.realpath(
        os.path.join(writer.outputdir, writer.filename)))
    if (not writer.force_overwrite
            and os.path.isfile(writer.filename)):
        raise ValueError(
            'Writing to an existing file without overwrite permission. '
            f'permission. Remove {writer.filename} or set '
            '"force_overwrite" in pipeline configuration for '
            f'{writer.name}')
    return writer


class Writer(PipelineItem):
    """Generic file writer.

    The job of any `Writer` in a `Pipeline` is to receive input
    returned by a previous `PipelineItem`, write its data to a
    particular file format, then return the same data unaltered so it
    can be used by a successive `PipelineItem`.

    :ivar filename: Name of file to write to.
    :type filename: str
    :ivar force_overwrite: Flag to allow data in `filename` to be
        overwritten if it already exists, defaults to `False`.
    :type force_overwrite: bool, optional
    :ivar remove: Flag to remove the dictionary from `data`,
        defaults to `False`.
    :type remove: bool, optional
    """
    filename: str
    force_overwrite: Optional[bool] = False
    remove: Optional[bool] = False

    _validate_filename = model_validator(mode='after')(validate_writer_model)

    def write(self, data):
        """Write the last `CHAP.pipeline.PipelineData` item in `data`
        as text to a file.

        :param data: Input data.
        :type data: list[CHAP.pipeline.PipelineData]
        :return: Contents of the input data.
        :rtype: list[PipelineData]
        """
        ddata = self.unwrap_pipelinedata(data)[-1]
        if os.path.isfile(self.filename) and not self.force_overwrite:
            raise FileExistsError(f'{self.filename} already exists')
        with open(self.filename, 'w') as f:
            f.write(ddata)
        if self.remove:
            data.pop()
        self.status = 'written' # Right now does nothing yet, but could
                                # add a sort of modification flag later
        return data


class OptionParser():
    """User based option parser."""
    def __init__(self):
        self.parser = argparse.ArgumentParser(prog='PROG')
        self.parser.add_argument(
            '--data', action='store',
            dest='data', default='', help='Input data')
        self.parser.add_argument(
            '--filename', action='store',
            dest='filename', default='', help='Output file')
        self.parser.add_argument(
            '--writer', action='store',
            dest='writer', default='Writer', help='Writer class name')
        self.parser.add_argument(
            '--log-level', choices=logging._nameToLevel.keys(),
            dest='log_level', default='INFO', help='logging level')


def main(opt_parser=OptionParser):
    """Main function."""
    optmgr = opt_parser()
    opts = optmgr.parser.parse_args()
    cls_name = opts.writer
    try:
        writer_cls = getattr(modules[__name__], cls_name)
    except AttributeError:
        print(f'Unsupported writer {cls_name}')
        raise

    writer = writer_cls()
    writer.logger.setLevel(getattr(logging, opts.log_level))
    log_handler = logging.StreamHandler()
    log_handler.setFormatter(logging.Formatter(
        '{name:20}: {message}', style='{'))
    writer.logger.addHandler(log_handler)
    data = writer.write(opts.data, opts.filename)
    print(f'Writer {writer} writes to {opts.filename}, data {data}')


if __name__ == '__main__':
    main()



"""Utils Pydantic model classes."""

# System modules
from typing import (
    Literal,
    Optional,
    Union,
)

# Third party imports
from pydantic import (
    Field,
    PrivateAttr,
    StrictBool,
    conint,
    conlist,
    confloat,
    constr,
    field_validator,
)
from typing_extensions import Annotated
import numpy as np

# Local modules
from CHAP.models import CHAPBaseModel
from CHAP.utils.general import not_zero, tiny

# pylint: disable=no-member
tiny = np.finfo(np.float64).resolution
# pylint: enable=no-member
s2pi = np.sqrt(2*np.pi)

#def constant(x, c=0.5):
def constant(x, c=0.0):
    """Return a linear function.

    constant(x, c) = c
    """
    return c*np.ones((x.size))


#def linear(x, slope=0.9, intercept=0.1):
def linear(x, slope=1.0, intercept=0.0):
    """Return a linear function.

    linear(x, slope, intercept) = slope * x + intercept
    """
    return slope * x + intercept


#def quadratic(x, a=0.5, b=0.4, c=0.1):
def quadratic(x, a=0.0, b=0.0, c=0.0):
    """Return a parabolic function.

    parabolic(x, a, b, c) = a * x**2 + b * x + c
    """
    return (a*x + b) * x + c


#def exponential(x, amplitude=1.0, decay=0.3):
def exponential(x, amplitude=1.0, decay=1.0):
    """Return an exponential function.

    exponential(x, amplitude, decay) = amplitude * exp(-x/decay)
    """
    return amplitude * np.exp(-x/not_zero(decay))


#def gaussian(x, amplitude=0.25, center=0.5, sigma=0.1):
def gaussian(x, amplitude=1.0, center=0.0, sigma=1.0):
    """Return a 1-dimensional Gaussian function.

    gaussian(x, amplitude, center, sigma) =
        (amplitude/(s2pi*sigma)) * exp(-(x-center)**2 / (2*sigma**2))
    """
    return ((amplitude/(max(tiny, s2pi*sigma)))
            * np.exp(-(x-center)**2 / max(tiny, (2*sigma**2))))


#def lorentzian(x, amplitude=0.3, center=0.5, sigma=0.1):
def lorentzian(x, amplitude=1.0, center=0.0, sigma=1.0):
    """Return a 1-dimensional Lorentzian function.

    lorentzian(x, amplitude, center, sigma) =
        (amplitude/(1 + ((1.0*x-center)/sigma)**2)) / (pi*sigma)
    """
    return ((amplitude/(1 + ((x-center)/max(tiny, sigma))**2))
            / max(tiny, (np.pi*sigma)))


def rectangle(
        x, amplitude=1.0, center1=0.0, sigma1=1.0, center2=1.0,
        sigma2=1.0, form='linear'):
    r"""
    Return a rectangle function.

    Starts at 0.0, rises to ``amplitude`` (at ``center1`` with width
    ``sigma1``), then drops to 0.0 (at ``center2`` with width
    ``sigma2``)

    :param x: Input values where the function is evaluated.
    :type x: float or numpy.ndarray
    :param amplitude: Maximum height of the rectangle, defaults to 1.0.
    :type amplitude: float, optional
    :param center1: Location of the rising edge, defaults to 0.0.
    :type center1: float, optional
    :param sigma1: Width or smoothness of the rising edge,
        defaults to 1.0.
    :type sigma1: float, optional
    :param center2: Location of the falling edge, defaults to 1.0.
    :type center2: float, optional
    :param sigma2: Width or smoothness of the falling edge,
        defaults to 1.0.
    :type sigma2: float, optional
    :param form: Shape type of the transition edges:

        - ``'linear'``: Simple ramp-up and ramp-down.
        - ``'atan'`` or ``'arctan'``: Inverse tangent transitions.
        - ``'erf'``: Error function (Gaussian-like) transitions.
        - ``'logistic'``: Sigmoidal (logistic function) transitions.
    :type form: str, optional

    :returns: The evaluated rectangle function values.
    :rtype: float or numpy.ndarray

    .. note::
        The output is calculated based on the selected ``form``:

        - **atan**: $\frac{A}{\pi} [ \arctan(arg_1) + \arctan(arg_2) ]$
        - **erf**:
          $\frac{1}{2} A [ \text{erf}(arg_1) + \text{erf}(arg_2) ]$
        - **logistic**:
          $A [ \frac{1}{1 + \exp(-arg_1)} +
          \frac{1}{1 + \exp(-arg_2)} - 1 ]$

        The function is constructed using normalized arguments for the
        rising and falling edges:
        $arg_1 = \frac{x - center_1}{\sigma_1}$
        and
        $arg_2 = \frac{center_2 - x}{\sigma_2}$
    """
    arg1 = (x - center1)/max(tiny, sigma1)
    arg2 = (center2 - x)/max(tiny, sigma2)

    if form == 'erf':
        # Third party modules
        # pylint: disable=no-name-in-module
        from scipy.special import erf

        rect = 0.5*(erf(arg1) + erf(arg2))
    elif form == 'logistic':
        rect = 1. - 1./(1. + np.exp(arg1)) - 1./(1. + np.exp(arg2))
    elif form in ('atan', 'arctan'):
        rect = (np.arctan(arg1) + np.arctan(arg2))/np.pi
    elif form == 'linear':
        rect = 0.5*(np.minimum(1, np.maximum(-1, arg1))
                   + np.minimum(1, np.maximum(-1, arg2)))
    else:
        raise ValueError(f'Invalid parameter form ({form})')

    return amplitude*rect


def validate_parameters(parameters, info):
    """Validate the parameters.

    :param parameters: Fit model parameters.
    :type parameters: list[FitParameter]
    :param info: Pydantic validator info object.
    :type info: pydantic_core._pydantic_core.ValidationInfo
    :return: List of fit model parameters.
    :rtype: list[FitParameter]
    """
    # System imports
    import inspect

    if 'model' in info.data:
        model = info.data['model']
    else:
        model = None
    if model is None or model == 'expression':
        return parameters
    sig = dict(inspect.signature(models[model]).parameters.items())
    sig.pop('x')

    # Check input model parameter validity
    for par in parameters:
        if par.name not in sig:
            raise ValueError('Invalid parameter {par.name} in {model} model')

    # Set model parameters
    output_parameters = []
    for sig_name, sig_par in sig.items():
        if model == 'rectangle' and sig_name == 'form':
            continue
        for par in parameters:
            if sig_name == par.name:
                break
        else:
            par = FitParameter(name=sig_name)
        if sig_par.default != sig_par.empty:
            par._default = sig_par.default
        output_parameters.append(par)

    return output_parameters


class FitParameter(CHAPBaseModel):
    """Class representing a specific fit parameter for the fit
    processor.
    """
    name: constr(strip_whitespace=True, min_length=1)
    value: Optional[confloat(allow_inf_nan=False)] = None
    min: Optional[confloat()] = -np.inf
    max: Optional[confloat()] = np.inf
    vary: StrictBool = True
    expr: Optional[constr(strip_whitespace=True, min_length=1)] = None
    _default: float = PrivateAttr()
    _init_value: float = PrivateAttr()
    _prefix: str = PrivateAttr()
    _stderr: float = PrivateAttr()

    @field_validator('min')
    @classmethod
    def validate_min(cls, value):
        """Validate the specified min.

        :param value: Field value to validate (`min`).
        :type value: Union[float, None]
        :return: Lower bound of fit parameter.
        :rtype: float
        """
        if value is None:
            return -np.inf
        return value

    @field_validator('max')
    @classmethod
    def validate_max(cls, value):
        """Validate the specified max.

        :param value: Field value to validate (`max`).
        :type value: Union[float, None]
        :return: Upper bound of fit parameter.
        :rtype: float
        """
        if value is None:
            return np.inf
        return value

    @property
    def default(self):
        """Return the _default attribute."""
        if hasattr(self, '_default'):
            return self._default
        return None

    @property
    def init_value(self):
        """Return the _init_value attribute."""
        if hasattr(self, '_init_value'):
            return self._init_value
        return None

    @property
    def prefix(self):
        """Return the _prefix attribute."""
        if hasattr(self, '_prefix'):
            return self._prefix
        return None

    @property
    def stderr(self):
        """Return the _stderr attribute."""
        if hasattr(self, '_stderr'):
            return self._stderr
        return None

    def set(self, value=None, min=None, max=None, vary=None, expr=None):
        """Set or update FitParameter attributes.

        :param value: Parameter value.
        :type value: float, optional
        :param min: Lower Parameter value bound. To remove the lower
            bound you must set min to `numpy.inf`.
        :type min: bool, optional
        :param max: Upper Parameter value bound. To remove the lower
            bound you must set max to `numpy.inf`.
        :type max: bool, optional
        :param vary: Whether the Parameter is varied during a fit.
        :type vary: bool, optional
        :param expr: Mathematical expression used to constrain the
            value during the fit. To remove a constraint you must
            supply an empty string.
        :type expr: str, optional
        """
        if expr is not None:
            if not isinstance(expr, str):
                raise ValueError(f'Invalid parameter expr ({expr})')
            if expr == '':
                expr = None
            self.expr = expr
            if expr is not None:
                self.value = None
                self.min = -np.inf
                self.max = np.inf
                self.vary = False
                return
        if min is not None:
            if not isinstance(min, (int, float)):
                raise ValueError(f'Invalid parameter min ({min})')
            self.min = min
        if max is not None:
            if not isinstance(max, (int, float)):
                raise ValueError(f'Invalid parameter max ({max})')
            self.max = max
        if vary is not None:
            if not isinstance(vary, bool):
                raise ValueError(f'Invalid parameter vary ({vary})')
            self.vary = vary
        if value is not None:
            if not isinstance(value, (int, float)):
                raise ValueError(f'Invalid parameter value ({value})')
            self.value = value
            if self.value > self.max:
                self.value = self.max
            elif self.value < self.min:
                self.value = self.min
            self.expr = None

class Constant(CHAPBaseModel):
    """Class representing a Constant model component.

    :ivar model: The model component base name (a prefix will be added
        if multiple identical model components are added).
    :type model: Literal['constant']
    :ivar parameters: Function parameters, defaults to those auto
        generated from the function signature (excluding the
        independent variable), defaults to `[]`.
    :type parameters: list[FitParameter], optional
    :ivar prefix: The model prefix, defaults to `''`.
    :type prefix: str, optional
    """
    model: Literal['constant']
    parameters: Annotated[
        conlist(item_type=FitParameter),
        Field(validate_default=True)] = []
    prefix: Optional[str] = ''

    _validate_parameters_parameters = field_validator(
        'parameters')(validate_parameters)


class Linear(CHAPBaseModel):
    """Class representing a Linear model component.

    :ivar model: The model component base name (a prefix will be added
        if multiple identical model components are added).
    :type model: Literal['linear']
    :ivar parameters: Function parameters, defaults to those auto
        generated from the function signature (excluding the
        independent variable), defaults to `[]`.
    :type parameters: list[FitParameter], optional
    :ivar prefix: The model prefix, defaults to `''`.
    :type prefix: str, optional
    """
    model: Literal['linear']
    parameters: Annotated[
        conlist(item_type=FitParameter),
        Field(validate_default=True)] = []
    prefix: Optional[str] = ''

    _validate_parameters_parameters = field_validator(
        'parameters')(validate_parameters)


class Quadratic(CHAPBaseModel):
    """Class representing a Quadratic model component.

    :ivar model: The model component base name (a prefix will be added
        if multiple identical model components are added).
    :type model: Literal['quadratic']
    :ivar parameters: Function parameters, defaults to those auto
        generated from the function signature (excluding the
        independent variable), defaults to `[]`.
    :type parameters: list[FitParameter], optional
    :ivar prefix: The model prefix, defaults to `''`.
    :type prefix: str, optional
    """
    model: Literal['quadratic']
    parameters: Annotated[
        conlist(item_type=FitParameter),
        Field(validate_default=True)] = []
    prefix: Optional[str] = ''

    _validate_parameters_parameters = field_validator(
        'parameters')(validate_parameters)


class Exponential(CHAPBaseModel):
    """Class representing an Exponential model component.

    :ivar model: The model component base name (a prefix will be added
        if multiple identical model components are added).
    :type model: Literal['exponential']
    :ivar parameters: Function parameters, defaults to those auto
        generated from the function signature (excluding the
        independent variable), defaults to `[]`.
    :type parameters: list[FitParameter], optional
    :ivar prefix: The model prefix, defaults to `''`.
    :type prefix: str, optional
    """
    model: Literal['exponential']
    parameters: Annotated[
        conlist(item_type=FitParameter),
        Field(validate_default=True)] = []
    prefix: Optional[str] = ''

    _validate_parameters_parameters = field_validator(
        'parameters')(validate_parameters)


class Gaussian(CHAPBaseModel):
    """Class representing a Gaussian model component.

    :ivar model: The model component base name (a prefix will be added
        if multiple identical model components are added).
    :type model: Literal['gaussian']
    :ivar parameters: Function parameters, defaults to those auto
        generated from the function signature (excluding the
        independent variable), defaults to `[]`.
    :type parameters: list[FitParameter], optional
    :ivar prefix: The model prefix, defaults to `''`.
    :type prefix: str, optional
    """
    model: Literal['gaussian']
    parameters: Annotated[
        conlist(item_type=FitParameter),
        Field(validate_default=True)] = []
    prefix: Optional[str] = ''

    _validate_parameters_parameters = field_validator(
        'parameters')(validate_parameters)


class Lorentzian(CHAPBaseModel):
    """Class representing a Lorentzian model component.

    :ivar model: The model component base name (a prefix will be added
        if multiple identical model components are added).
    :type model: Literal['lorentzian']
    :ivar parameters: Function parameters, defaults to those auto
        generated from the function signature (excluding the
        independent variable), defaults to `[]`.
    :type parameters: list[FitParameter], optional
    :ivar prefix: The model prefix, defaults to `''`.
    :type prefix: str, optional
    """
    model: Literal['lorentzian']
    parameters: Annotated[
        conlist(item_type=FitParameter),
        Field(validate_default=True)] = []
    prefix: Optional[str] = ''

    _validate_parameters_parameters = field_validator(
        'parameters')(validate_parameters)


class Rectangle(CHAPBaseModel):
    """Class representing a Rectangle model component.

    :ivar model: The model component base name (a prefix will be added
        if multiple identical model components are added).
    :type model: Literal['rectangle']
    :ivar parameters: Function parameters, defaults to those auto
        generated from the function signature (excluding the
        independent variable), defaults to `[]`.
    :type parameters: list[FitParameter], optional
    :ivar prefix: The model prefix, defaults to `''`.
    :type prefix: str, optional
    """
    model: Literal['rectangle']
    parameters: Annotated[
        conlist(item_type=FitParameter),
        Field(validate_default=True)] = []
    prefix: Optional[str] = ''

    _validate_parameters_parameters = field_validator(
        'parameters')(validate_parameters)


class Expression(CHAPBaseModel):
    """Class representing an Expression model component.

    :ivar model: The model component base name (a prefix will be added
        if multiple identical model components are added).
    :type model: Literal['expression']
    :ivar expr: Mathematical expression to represent the model
        component.
    :type expr: str
    :ivar parameters: Function parameters, defaults to those auto
        generated from the model expression (excluding the
        independent variable), defaults to `[]`.
    :type parameters: list[FitParameter], optional
    :ivar prefix: The model prefix, defaults to `''`.
    :type prefix: str, optional
    """
    model: Literal['expression']
    expr: constr(strip_whitespace=True, min_length=1)
    parameters: Annotated[
        conlist(item_type=FitParameter),
        Field(validate_default=True)] = []
    prefix: Optional[str] = ''

    _validate_parameters_parameters = field_validator(
        'parameters')(validate_parameters)


class Multipeak(CHAPBaseModel):
    """Class representing a multipeak model.

    :ivar model: The model component base name (a prefix will be added
        if multiple identical model components are added).
    :type model: Literal['expression']
    :ivar centers: Peak centers.
    :type center: list[float]
    :ivar centers_range: Range of peak centers around their centers.
    :type centers_range: float, optional
    :ivar fit_type: Type of fit, defaults to `'unconstrained'`.
    :type fit_type: Literal['uniform', 'unconstrained'], optional.
    :ivar fwhm_min: Lower limit of the fwhm of the peaks.
    :type fwhm_min: float, optional
    :ivar fwhm_max: Upper limit of the fwhm of the peaks.
    :type fwhm_max: float, optional
    :ivar peak_models: Type of peaks, defaults to `'gaussian'`.
    :type peak_models: Literal['gaussian', 'lorentzian'], optional.
    """
    model: Literal['multipeak']
    centers: conlist(item_type=confloat(allow_inf_nan=False), min_length=1)
    centers_range: Optional[confloat(allow_inf_nan=False)] = None
    fit_type: Optional[Literal['uniform', 'unconstrained']] = 'unconstrained'
    fwhm_min: Optional[confloat(allow_inf_nan=False)] = None
    fwhm_max: Optional[confloat(allow_inf_nan=False)] = None
    peak_models: Literal['gaussian', 'lorentzian'] = 'gaussian'


models = {
    'constant': constant,
    'linear': linear,
    'quadratic': quadratic,
    'exponential': exponential,
    'gaussian': gaussian,
    'lorentzian': lorentzian,
    'rectangle': rectangle,
}

model_classes = (
    Constant,
    Linear,
    Quadratic,
    Exponential,
    Gaussian,
    Lorentzian,
    Rectangle,
)


class FitConfig(CHAPBaseModel):
    """Class representing the configuration for the fit processor.

    :ivar code: Specifies is lmfit is used to perform the fit or if
        the scipy fit method is called directly, default to `'lmfit'`.
    :type code: Literal['lmfit', 'scipy'], optional
    :ivar parameters: Fit model parameters in addition to those
        implicitly defined through the build-in model functions,
        defaults to `[]`'
    :type parameters: list[FitParameter], optional
    :ivar models: The component(s) of the (composite) fit model.
    :type models: Union[Constant, Linear, Quadratic, Exponential,
        Gaussian, Lorentzian, Rectangle, Expression, Multipeak]
    :ivar rel_height_cutoff: Relative peak height cutoff for
        peak fitting (any peak with a height smaller than
        `rel_height_cutoff` times the maximum height of all peaks 
        gets removed from the fit model).
    :type rel_height_cutoff: float, optional
    :ivar num_proc: The number of processors used in fitting a map
        of data, defaults to `1`.
    :type num_proc: int, optional
    :ivar plot: Weather a plot of the fit result is generated,
        defaults to `False`.
    :type plot: bool, optional.
    :ivar print_report:  Weather to generate a fit result printout,
        defaults to `False`.
    :type print_report: bool, optional.
    """
    code: Literal['lmfit', 'scipy'] = 'scipy'
    parameters: conlist(item_type=FitParameter) = []
    models: conlist(item_type=Union[
        Constant, Linear, Quadratic, Exponential, Gaussian, Lorentzian,
        Rectangle, Expression, Multipeak], min_length=1)
    method: Literal[
        'leastsq', 'trf', 'dogbox', 'lm', 'least_squares'] = 'leastsq'
    rel_height_cutoff: Optional[
        confloat(gt=0, lt=1.0, allow_inf_nan=False)] = None
    num_proc: conint(gt=0) = 1
    plot: StrictBool = False
    print_report:  StrictBool = False
    memfolder: str = 'joblib_memmap'

    @field_validator('method')
    @classmethod
    def validate_method(cls, method, info):
        """Validate the specified method.

        :param method: The value of `method` to validate.
        :type method: str
        :param info: Pydantic validator info object.
        :type info: pydantic_core._pydantic_core.ValidationInfo
        :return: Fit method.
        :rtype: str
        """
        code = info.data['code']
        if code == 'lmfit':
            if method not in ('leastsq', 'least_squares'):
                method = 'leastsq'
        elif method == 'least_squares':
            method = 'leastsq'

        return method



"""This subpackage contains `PipelineItems` to communicate with FOXDEN
services.
"""

from CHAP.foxden.processor import (
#    FoxdenMetadataProcessor,
    FoxdenProvenanceProcessor,
)
from CHAP.foxden.reader import (
    FoxdenDataDiscoveryReader,
    FoxdenMetadataReader,
    FoxdenProvenanceReader,
    FoxdenSpecScansReader,
)
from CHAP.foxden.writer import (
    FoxdenDoiWriter,
    FoxdenMetadataWriter,
    FoxdenProvenanceWriter,
)



"""PipelineItems for interacting with NeXus file objects"""

from CHAP import Processor


class NexusMakeLinkProcessor(Processor):
    """Processor to run
    [`nexusformat.nexus.tree.NXgroup.makelink`](https://nexpy.github.io/nexpy/treeapi.html#nexusformat.nexus.tree.NXgroup.makelink)
    within a given `NXroot`"""
    def process(self, data, link_from, link_to,
                nxname=None, abspath=False):
        """Create links between Nexus objects within the given
        PipelineData.

        This method takes an `NXroot` and creates links from the
        objects specified in `link_from` to those in `link_to`. If the
        underlying file is read-only, a copy is made before
        modifying. Returns the modified `NXroot` object containing
        both the targets and their linked counterparts.

        :param data: Data from previous PipelineItems.
        :type data: list[PipelineData]

        :param link_from: Path(s) within the NXroot whose objects
            should be linked. Can be a single path (str) or a list of
            paths.
        :type link_from: str | list[str]

        :param link_to: Path(s) within the NXroot that serve as link
            targets.  Can be a single path (str) or a list of paths.
        :type link_to: str | list[str]

        :param nxname: Optional name to assign to the created link. If
            None, the default naming rules from `makelink` are
            applied.
        :type nxname: str | None

        :param abspath: Whether to create an absolute link path (True)
            or a relative one (False). Defaults to False.
        :type abspath: bool

        :returns: The modified NXroot object containing the new links.
        :rtype: NXroot
        """
        from CHAP.utils.general import nxcopy
        
        root = self.get_data(data)
        self.logger.debug(f'root.nxfile.mode = {root.nxfile.mode}')
        if root.nxfile.mode == 'r':
            # root belongs to a readonly file, copy to proceed.
            root = nxcopy(root)

        if isinstance(link_from, str):
            link_from = [link_from]
        if isinstance(link_to, str):
            link_to = [link_to]

        for _from in link_from:
            for _to in link_to:
                origin = root[_from]
                target = root[_to]
                self.logger.debug(f'linking to {_to} from {_from}')
                origin.makelink(target, name=nxname, abspath=abspath)

        return root



"""EDD Pydantic model classes."""

# System modules
from copy import deepcopy
import os
from typing import (
    Literal,
    Optional,
    Union,
)

# Third party modules
import numpy as np
from hexrd.material import Material
#from CHAP.utils.material import Material
from pydantic import (
    Field,
    FilePath,
    PrivateAttr,
    confloat,
    conint,
    conlist,
    constr,
    field_validator,
    model_validator,
)
from scipy.interpolate import interp1d
from typing_extensions import Annotated

# Local modules
from CHAP.models import CHAPBaseModel
from CHAP.common.models.map import Detector
from CHAP.utils.models import Multipeak
#from CHAP.utils.parfile import ParFile


# Baseline configuration class

class BaselineConfig(CHAPBaseModel):
    """Baseline model configuration class.

    :ivar lam: The &lambda (smoothness) parameter (the balance
        between the residual of the data and the baseline and the
        smoothness of the baseline). The suggested range is between
        100 and 10^8, defaults to `10^6`.
    :type lam: float, optional
    :ivar max_iter: The maximum number of iterations,
        defaults to `100`.
    :type max_iter: int, optional
    :ivar tol: The convergence tolerence, defaults to `1.e-6`.
    :type tol: float, optional
    """
    attrs: Optional[dict] = {}
    lam: confloat(gt=0, allow_inf_nan=False) = 1.e6
    max_iter: conint(gt=0) = 100
    tol: confloat(gt=0, allow_inf_nan=False) = 1.e-6


# Fit configuration class

class FitConfig(CHAPBaseModel):
    """Fit parameters configuration class for peak fitting.

    :ivar background: Background model for peak fitting, defaults
        to `constant`.
    :type background: str, list[str], optional
    :ivar baseline: Automated baseline subtraction configuration,
        defaults to `False`.
    :type baseline: Union(bool, BaselineConfig), optional
    :ivar centers_range: Peak centers range for peak fitting.
        The allowed range for the peak centers will be the initial
        values &pm; `centers_range` (in MCA channels for calibration
        or keV for strain analysis). Defaults to `20` for calibration
        and `2.0` for strain analysis.
    :type centers_range: float, optional
    :ivar energy_mask_ranges: List of MCA energy mask ranges in keV
        for selecting the data to be included after applying a mask
        (bounds are inclusive). Specify either energy_mask_ranges or
        mask_ranges, not both.
    :type energy_mask_ranges: list[[float, float]], optional
    :ivar fwhm_min: Minimum FWHM for peak fitting (in MCA channels
        for calibration or keV for strain analysis). Defaults to `3`
        for calibration and `0.25` for strain analysis.
    :type fwhm_min: float, optional
    :ivar fwhm_max: Maximum FWHM for peak fitting (in MCA channels
        for calibration or keV for strain analysis). Defaults to `25`
        for calibration and `2.0` for strain analysis.
    :type fwhm_max: float, optional
    :ivar mask_ranges: List of MCA channel bin ranges for selecting
        the data to be included in the energy calibration after
        applying a mask (bounds are inclusive). Specify for
        energy calibration only.
    :type mask_ranges: list[[int, int]], optional
    :ivar backgroundpeaks: Additional background peaks (their
        associated fit parameters in units of keV).
    :type backgroundpeaks: CHAP.utils.models.Multipeak, optional
    """
    background: Optional[conlist(item_type=constr(
        strict=True, strip_whitespace=True, to_lower=True))] = ['constant']
    baseline: Optional[Union[bool, BaselineConfig]] = None
    centers_range: Optional[confloat(gt=0, allow_inf_nan=False)] = 20
    energy_mask_ranges: Optional[conlist(
        min_length=1,
        item_type=conlist(
            min_length=2,
            max_length=2,
            item_type=confloat(allow_inf_nan=False)))] = None
    fwhm_min: Optional[confloat(gt=0, allow_inf_nan=False)] = 3
    fwhm_max: Optional[confloat(gt=0, allow_inf_nan=False)] = 25
    mask_ranges: Optional[conlist(
        min_length=1,
        item_type=conlist(
            min_length=2,
            max_length=2,
            item_type=conint(ge=0)))] = None
    backgroundpeaks: Optional[Multipeak] = None

    @field_validator('background', mode='before')
    @classmethod
    def validate_background(cls, background):
        """Validate the background model.

        :ivar background: Background model for peak fitting.
        :type background: str, list[str], optional
        :return: List of validated background models.
        :rtype: list[str]
        """
        if background is None:
            return background
        if isinstance(background, str):
            return [background]
        return sorted(background)

    @field_validator('baseline', mode='before')
    @classmethod
    def validate_baseline(cls, baseline):
        """Validate the baseline configuration.

        :ivar baseline: Automated baseline subtraction configuration.
        :type baseline: Union(bool, BaselineConfig), optional
        :return: Validated baseline subtraction configuration.
        :rtype: bool, BaselineConfig
        """
        if isinstance(baseline, bool) and baseline:
            return BaselineConfig()
        return baseline

    @field_validator('energy_mask_ranges', mode='before')
    @classmethod
    def validate_energy_mask_ranges(cls, energy_mask_ranges):
        """Validate the mask ranges for selecting the data to include.

        :ivar energy_mask_ranges: List of MCA energy mask ranges in keV
            for selecting the data to be included after applying a mask
            (bounds are inclusive).
        :type energy_mask_ranges: list[[float, float]], optional
        :return: Validated energy mask ranges.
        :rtype: list[[float, float]]
        """
        if energy_mask_ranges:
            return sorted([sorted(v) for v in energy_mask_ranges])
        return energy_mask_ranges

    @field_validator('mask_ranges', mode='before')
    @classmethod
    def validate_mask_ranges(cls, mask_ranges):
        """Validate the mask ranges for selecting the data to include.

        :ivar mask_ranges: List of MCA channel bin ranges for selecting
            the data to be included after applying a mask
            (bounds are inclusive).
        :type mask_ranges: list[[int, int]], optional
        :return: Validated mask ranges.
        :rtype: list[[int, int]]
        """
        if mask_ranges:
            return sorted([sorted(v) for v in mask_ranges])
        return mask_ranges


# Material configuration class

class MaterialConfig(CHAPBaseModel):
    """Sample material parameters configuration class.

    :ivar material_name: Sample material name.
    :type material_name: str, optional
    :ivar lattice_parameters: Lattice spacing(s) in angstroms.
    :type lattice_parameters: float, list[float], optional
    :ivar sgnum: Space group of the material.
    :type sgnum: int, optional
    """
    #RV FIX create a getter for lattice_parameters that always returns a list?
    material_name: Optional[constr(strip_whitespace=True, min_length=1)] = None
    lattice_parameters: Optional[Union[
        confloat(gt=0, allow_inf_nan=False),
        conlist(
            min_length=1, max_length=6,
            item_type=confloat(gt=0, allow_inf_nan=False))]] = None
    sgnum: Optional[conint(ge=0)] = None

    _material: Optional[Material]

    @model_validator(mode='after')
    def validate_materialconfig_after(self):
        """Create and validate the private attribute _material.

        :return: The validated list of class properties.
        :rtype: MaterialConfig
        """
        # Local modules
        from CHAP.edd.utils import make_material
#        from CHAP.utils.material import Material

        self._material = make_material(
            self.material_name, self.sgnum, self.lattice_parameters)
#        self._material = Material.make_material(
#            self.material_name, sgnum=self.sgnum,
#            lattice_parameters_angstroms=self.lattice_parameters,
#            pos=['4a', '8c'])
            #pos=[(0,0,0), (1/4, 1/4, 1/4), (3/4, 3/4, 3/4)])
        self.lattice_parameters = list([
            x.getVal('angstrom') if x.isLength() else x.getVal('radians')
            for x in self._material._lparms])
        return self


# Detector configuration classes

class MCADetectorCalibration(Detector, FitConfig):
    """Class representing metadata required to configure a single MCA
    detector element to perform detector calibration.

    :ivar energy_calibration_coeffs: Detector channel index to energy
        polynomial conversion coefficients ([a, b, c] with
        E_i = a*i^2 + b*i + c).
    :type energy_calibration_coeffs:
        list[float, float, float], optional
    :ivar num_bins: Number of MCA channels.
    :type num_bins: int, optional
    :ivar tth_max: Detector rotation about lab frame x axis.
    :type tth_max: float, optional
    :ivar tth_tol: Minimum resolvable difference in 2&theta between
        two unique Bragg peaks,
    :type tth_tol: float, optional
    :ivar tth_calibrated: Calibrated value for 2&theta.
    :type tth_calibrated: float, optional
    :ivar tth_initial_guess: Initial guess for 2&theta superseding
        the global one in MCATthCalibrationConfig.
    :type tth_initial_guess: float, optional
    """
    processor_type: Literal['calibration']
    energy_calibration_coeffs: Optional[conlist(
        min_length=3, max_length=3,
        item_type=confloat(allow_inf_nan=False))] = None
    num_bins: Optional[conint(gt=0)] = None
    tth_max: Optional[confloat(gt=0, allow_inf_nan=False)] = None
    tth_tol: Optional[confloat(gt=0, allow_inf_nan=False)] = None
    tth_calibrated: Optional[confloat(gt=0, allow_inf_nan=False)] = None
    tth_initial_guess: Optional[confloat(gt=0, allow_inf_nan=False)] = None

    _energy_calibration_mask_ranges: conlist(
        min_length=1,
        item_type=conlist(
            min_length=2,
            max_length=2,
            item_type=conint(ge=0))) = PrivateAttr()
    _hkl_indices: list = PrivateAttr()

#    def add_calibration(self, calibration):
#        """Finalize values for some fields using a calibration
#        MCADetectorCalibration corresponding to the same detector.
#
#        :param calibration: Existing calibration configuration.
#        :type calibration: MCADetectorCalibration
#        """
#        raise RuntimeError('To do')
#        for field in ['energy_calibration_coeffs', 'num_bins',
#                      '_energy_calibration_mask_ranges']:
#            setattr(self, field, deepcopy(getattr(calibration, field)))
#        if self.tth_calibrated is not None:
#            self.logger.warning(
#                'Ignoring tth_calibrated in calibration configuration')
#            self.tth_calibrated = None

    @property
    def energies(self):
        """Return calibrated bin energies."""
        a, b, c = tuple(self.energy_calibration_coeffs)
        channel_bins = np.arange(self.num_bins)
        return (a*channel_bins + b)*channel_bins + c

    @property
    def hkl_indices(self):
        """Return the hkl_indices consistent with the selected energy
        ranges (include_energy_ranges).
        """
        if hasattr(self, '_hkl_indices'):
            return self._hkl_indices
        return []

    @hkl_indices.setter
    def hkl_indices(self, value):
        """Set the private attribute `hkl_indices`."""
        self._hkl_indices = value

    def convert_mask_ranges(self, mask_ranges):
        """Given a list of mask ranges in channel bins, set the
        corresponding list of channel energy mask ranges.

        :param mask_ranges: A list of mask ranges to
            convert to energy mask ranges.
        :type mask_ranges: list[[int,int]]
        """
        energies = self.energies
        self.energy_mask_ranges = [
            [float(energies[i]) for i in range_]
             for range_ in sorted([sorted(v) for v in mask_ranges])]

    def get_mask_ranges(self):
        """Return the value of `mask_ranges` if set or convert the
        `energy_mask_ranges` from channel energies to channel indices.
        """
        if self.mask_ranges:
            return self.mask_ranges
        if self.energy_mask_ranges is None:
            return None

        # Local modules
        from CHAP.utils.general import (
            index_nearest_down,
            index_nearest_up,
        )

        mask_ranges = []
        energies = self.energies
        for e_min, e_max in self.energy_mask_ranges:
            mask_ranges.append(
                [index_nearest_down(energies, e_min),
                 index_nearest_up(energies, e_max)])
        return mask_ranges

    def mca_mask(self):
        """Get a boolean mask array to use on this MCA element's data.
        Note that the bounds of the mask ranges are inclusive.

        :return: Boolean mask array.
        :rtype: numpy.ndarray
        """
        mask = np.asarray([False] * self.num_bins)
        mask_ranges = self.get_mask_ranges()
        channel_bins = np.arange(self.num_bins, dtype=np.int32)
        for (min_, max_) in mask_ranges:
            mask = np.logical_or(
                mask,
                np.logical_and(channel_bins >= min_, channel_bins <= max_))
        return mask

    def set_energy_calibration_mask_ranges(self):
        self._energy_calibration_mask_ranges = deepcopy(self.mask_ranges)


class MCADetectorDiffractionVolumeLength(MCADetectorCalibration):
    """Class representing metadata required to perform a diffraction
    volume length measurement for a single MCA detector element.

    :ivar dvl: Measured diffraction volume length.
    :type dvl: float, optional
    :ivar fit_amplitude: Amplitude of the Gaussian fit.
    :type fit_amplitude: float, optional
    :ivar fit_center: Center of the Gaussian fit.
    :type fit_center: float, optional
    :ivar fit_sigma: Sigma of the Gaussian fit.
    :type fit_sigma: float, optional
    """
    processor_type: Literal['diffractionvolumelength']
    dvl: Optional[confloat(gt=0, allow_inf_nan=False)] = None
    fit_amplitude: Optional[float] = None
    fit_center: Optional[float] = None
    fit_sigma: Optional[float] = None


class MCADetectorStrainAnalysis(MCADetectorCalibration):
    """Class representing metadata required to perform a strain
    analysis.

    :ivar centers_range: Peak centers range for peak fitting.
        The allowed range for the peak centers will be the initial
        values &pm; `centers_range` (in keV), defaults to `2.0`.
    :type centers_range: float, optional
    :ivar fwhm_min: Minimum FWHM for peak fitting (in keV),
        defaults to `0.25`.
    :type fwhm_min: float, optional
    :ivar fwhm_max: Maximum FWHM for peak fitting (in keV),
        defaults to `2.0`.
    :ivar peak_models: Peak model for peak fitting,
        defaults to `'gaussian'`.
    :type peak_models: Literal['gaussian', 'lorentzian']],
        list[Literal['gaussian', 'lorentzian']]], optional
    :ivar rel_height_cutoff: Relative peak height cutoff for
        peak fitting (any peak with a height smaller than
        `rel_height_cutoff` times the maximum height of all peaks 
        gets removed from the fit model), defaults to `None`.
    :type rel_height_cutoff: float, optional
    :ivar tth_file: Path to the file with the 2&theta map.
    :type tth_file: FilePath, optional
    :ivar tth_map: Map of the 2&theta values.
    :type tth_map: numpy.ndarray, optional
    """
    centers_range: Optional[confloat(gt=0, allow_inf_nan=False)] = 2
    fwhm_min: Optional[confloat(gt=0, allow_inf_nan=False)] = 0.25
    fwhm_max: Optional[confloat(gt=0, allow_inf_nan=False)] = 2.0
    processor_type: Literal['strainanalysis']
    peak_models: Union[
        conlist(min_length=1, item_type=Literal['gaussian', 'lorentzian']),
        Literal['gaussian', 'lorentzian']] = 'gaussian'
    rel_height_cutoff: Optional[
        confloat(gt=0, lt=1.0, allow_inf_nan=False)] = None
#    tth_file: Optional[FilePath] = None
#    tth_map: Optional[np.ndarray] = None

    _calibration_energy_mask_ranges: conlist(
        min_length=1,
        item_type=conlist(
            min_length=2,
            max_length=2,
            item_type=confloat(allow_inf_nan=False))) = PrivateAttr()

    def add_calibration(self, calibration):
        """Finalize values for some fields using a tth calibration
        MCADetectorStrainAnalysis corresponding to the same detector.

        :param calibration: Existing calibration configuration to use
            by MCAElementStrainAnalysisConfig.
        :type calibration: MCADetectorStrainAnalysis
        """
        for field in ['energy_calibration_coeffs', 'num_bins',
                      'tth_calibrated']:
            setattr(self, field, deepcopy(getattr(calibration, field)))
        if self.energy_mask_ranges is None:
            self.energy_mask_ranges = deepcopy(calibration.energy_mask_ranges)
        self._calibration_energy_mask_ranges = deepcopy(
            calibration.energy_mask_ranges)

    def get_calibration_mask_ranges(self):
        """Return the `_calibration_energy_mask_ranges` converted from
        channel energies to channel indices.
        """
        if not hasattr(self, '_calibration_energy_mask_ranges'):
            return None

        # Local modules
        from CHAP.utils.general import (
            index_nearest_down,
            index_nearest_up,
        )

        energy_mask_ranges = []
        energies = self.energies
        for e_min, e_max in self._calibration_energy_mask_ranges:
            energy_mask_ranges.append(
                [index_nearest_down(energies, e_min),
                 index_nearest_up(energies, e_max)])
        return energy_mask_ranges

    def get_tth_map(self, map_shape):
        """Return the map of 2&theta values to use -- may vary at each
        point in the map.

        :param map_shape: The shape of the suplied 2&theta map.
        :return: Map of 2&theta values.
        :rtype: numpy.ndarray
        """
        if getattr(self, 'tth_map', None) is not None:
            if self.tth_map.shape != map_shape:
                raise ValueError(
                    'Invalid "tth_map" field shape '
                    f'{self.tth_map.shape} (expected {map_shape})')
            return self.tth_map
        return np.full(map_shape, self.tth_calibrated)


MCADetector = Annotated[
    Union[
        MCADetectorCalibration,
        MCADetectorDiffractionVolumeLength,
        MCADetectorStrainAnalysis],
    Field(discriminator='processor_type')
]


class MCADetectorConfig(FitConfig):
    """Class representing metadata required to configure a full MCA
    detector.

    :ivar detectors: List of individual MCA detector elements.
    :type detectors: list[MCADetector], optional
    """
    processor_type: Literal[
        'calibration', 'diffractionvolumelength', 'strainanalysis']
    detectors: Optional[conlist(min_length=1, item_type=MCADetector)] = []

    _exclude = set(vars(FitConfig()).keys())

    @model_validator(mode='before')
    @classmethod
    def validate_mcadetectorconfig_before(cls, data):
        if isinstance(data, dict):
            processor_type = data.get('processor_type').lower()
            if 'detectors' in data:
                detectors = data.pop('detectors')
                for d in detectors:
                    d['processor_type'] = processor_type
                    attrs = d.pop('attrs', {})
                    if 'default_fields' in attrs:
                        attrs.pop('default_fields')
                    if attrs:
                        d['attrs'] = attrs
                data['detectors'] = detectors
        return data

    @model_validator(mode='after')
    def validate_mcadetectorconfig_after(self):
        if self.detectors:
            self.update_detectors()
        return self

    def update_detectors(self):
        """Update individual detector parameters with any non-default
        values from the global detector configuration.
        """
        for k, v in self:
            if k in self.model_fields_set:
                for d in self.detectors:
                    if hasattr(d, k):
                        setattr(d, k, deepcopy(v))


# Processor configuration classes

class DiffractionVolumeLengthConfig(FitConfig):
    """Class representing metadata required to perform a diffraction
    volume length calculation for an EDD setup using a steel-foil
    raster scan.

    :ivar max_energy_kev: Maximum channel energy of the MCA in
        keV, defaults to `200.0`.
    :type max_energy_kev: float, optional
    :ivar measurement_mode: Placeholder for recording whether the
        measured DVL value was obtained through the automated
        calculation or a manual selection, defaults to `'auto'`.
    :type measurement_mode: Literal['manual', 'auto'], optional
    :ivar sample_thickness: Thickness of scanned foil sample. Quantity
        must be provided in the same units as the values of the
        scanning motor.
    :type sample_thickness: float
    :ivar sigma_to_dvl_factor: The DVL is obtained by fitting a reduced
        form of the MCA detector data. `sigma_to_dvl_factor` is a
        scalar value that converts the standard deviation of the
        gaussian fit to the measured DVL, defaults to `3.5`.
    :type sigma_to_dvl_factor: Literal[2.0, 3.5, 4.0], optional
    """
    max_energy_kev: Optional[confloat(gt=0, allow_inf_nan=False)] = 200.0
    measurement_mode: Optional[Literal['manual', 'auto']] = 'auto'
    sample_thickness: Optional[confloat(gt=0, allow_inf_nan=False)] = None
    sigma_to_dvl_factor: Optional[Literal[2.0, 3.5, 4.0]] = 3.5

    _exclude = set(vars(FitConfig()).keys())

    @model_validator(mode='after')
    def validate_diffractionvolumelengthconfig_after(self):
        """Update the configuration with costum defaults after the
        normal native pydantic validation.

        :return: Updated energy calibration configuration class.
        :rtype: DiffractionVolumeLengthConfig
        """
        if self.measurement_mode == 'manual':
            self._exclude |= {'sigma_to_dvl_factor'}
        return self


class MCACalibrationConfig(CHAPBaseModel):
    """Base class representing metadata required to perform an energy
    or 2&theta calibration of an MCA detector.

    :ivar flux_file: File name of the csv flux file containing station
        beam energy in eV (column 0) versus flux (column 1).
    :type flux_file: str, optional
    :ivar materials: Material configurations for the calibration,
        defaults to [`Ceria`].
    :type materials: list[MaterialConfig], optional
    :ivar peak_energies: Theoretical locations of the fluorescence
        peaks in keV to use for calibrating the MCA channel energies.
    :type peak_energies: list[float], optional for energy calibration
    :ivar scan_step_indices: Optional scan step indices to use for the
        calibration. If not specified, the calibration will be
        performed on the average of all MCA spectra for the scan.
    :type scan_step_indices: int, str, list[int], optional

    Note: Fluorescence data:
        https://physics.nist.gov/PhysRefData/XrayTrans/Html/search.html
    """
    flux_file: Optional[FilePath] = None
    materials: Optional[conlist(item_type=MaterialConfig)] = [MaterialConfig(
        material_name='CeO2', lattice_parameters=5.41153, sgnum=225)]
    peak_energies: Optional[conlist(
        min_length=2, item_type=confloat(gt=0, allow_inf_nan=False))] = [
            34.279, 34.720, 39.258, 40.233]
    scan_step_indices: Optional[
        conlist(min_length=1, item_type=conint(ge=0))] = None

    @model_validator(mode='before')
    @classmethod
    def validate_mcacalibrationconfig_before(cls, data):
        """Ensure that a valid configuration was provided and finalize
        flux_file filepath.

        :param data: Pydantic validator data object.
        :type data: MCACalibrationConfig,
            pydantic_core._pydantic_core.ValidationInfo
        :return: The currently validated list of class properties.
        :rtype: dict
        """
        if isinstance(data, dict):
            inputdir = data.get('inputdir')
            if inputdir is not None:
                flux_file = data.get('flux_file')
                if flux_file is not None and not os.path.isabs(flux_file):
                    data['flux_file'] = os.path.join(inputdir, flux_file)
        return data

    @field_validator('scan_step_indices', mode='before')
    @classmethod
    def validate_scan_step_indices(cls, scan_step_indices):
        """Validate the specified list of scan numbers.

        :ivar scan_step_indices: Optional scan step indices to use for
            the calibration. If not specified, the calibration will be
            performed on the average of all MCA spectra for the scan.
        :type scan_step_indices: int, str, list[int], optional
        :raises ValueError: Invalid experiment type.
        :return: List of step indices.
        :rtype: list[int]
        """
        if isinstance(scan_step_indices, int):
            scan_step_indices = [scan_step_indices]
        elif isinstance(scan_step_indices, str):
            # Local modules
            from CHAP.utils.general import string_to_list

            scan_step_indices = string_to_list(scan_step_indices)
        return scan_step_indices

    def flux_file_energy_range(self):
        """Get the energy range in the flux correction file.

        :return: The energy range in the flux correction file.
        :rtype: tuple(float, float)
        """
        if self.flux_file is None:
            return None
        flux = np.loadtxt(self.flux_file)
        energies = flux[:,0]/1.e3
        return energies.min(), energies.max()

    def flux_correction_interpolation_function(self):
        """Get an interpolation function to correct MCA data for the
        relative energy flux of the incident beam.

        :return: Energy flux correction interpolation function.
        :rtype: scipy.interpolate._polyint._Interpolator1D
        """
        if self.flux_file is None:
            return None
        flux = np.loadtxt(self.flux_file)
        energies = flux[:,0]/1.e3
        relative_intensities = flux[:,1]/np.max(flux[:,1])
        interpolation_function = interp1d(energies, relative_intensities)
        return interpolation_function


class MCAEnergyCalibrationConfig(MCACalibrationConfig):
    """Base class representing metadata required to perform an energy
    calibration of an MCA detector.

    :ivar max_energy_kev: Maximum channel energy of the MCA in
        keV, defaults to `200.0`.
    :type max_energy_kev: float, optional
    :ivar max_peak_index: Index of the peak in `peak_energies`
        with the highest amplitude, defaults to `1` (the second peak)
        for CeO2 calibration. Required for any other materials.
    :type max_peak_index: int, optional
    """
    max_energy_kev: Optional[confloat(gt=0, allow_inf_nan=False)] = 200.0
    max_peak_index: Optional[conint(ge=0)] = None

    @model_validator(mode='before')
    @classmethod
    def validate_mcaenergycalibrationconfig_before(cls, data):
        if isinstance(data, dict):
            detectors = data.pop('detectors', None)
            if detectors is not None:
                data['detector_config'] = {'detectors': detectors}
        return data

    @model_validator(mode='after')
    def validate_mcaenergycalibrationconfig_after(self):
        """Validate the detector (energy) mask ranges and update any
        detector configuration parameters not superseded by their
        individual values.

        :return: Updated energy calibration configuration class.
        :rtype: MCAEnergyCalibrationConfig
        """
        if self.peak_energies is None:
            raise ValueError('peak_energies is required')
        if (self.max_peak_index is not None
                and not 0 <= self.max_peak_index < len(self.peak_energies)):
            raise ValueError('max_peak_index out of bounds')
        return self

    @field_validator('max_peak_index', mode='before')
    @classmethod
    def validate_max_peak_index(cls, max_peak_index, info):
        if max_peak_index is None:
            materials = info.data.get('materials', [])
            if len(materials) != 1 or materials[0].material_name != 'CeO2':
                raise ValueError('max_peak_index is required unless the '
                                 'calibration material is CeO2')
            max_peak_index = 1
        return max_peak_index

class MCATthCalibrationConfig(MCACalibrationConfig):
    """Class representing metadata required to perform a 2&theta
    calibration of an MCA detector.

    :ivar calibration_method: Type of calibration method,
        defaults to `'direct_fit_bragg'`.
    :type calibration_method:
        Literal['direct_fit_bragg', 'direct_fit_tth_ecc'], optional
    :ivar detectors: List of individual MCA detector element
        calibration configurations.
    :ivar quadratic_energy_calibration: Adds a quadratic term to
        the detector channel index to energy conversion, defaults
        to `False` (linear only).
    :type quadratic_energy_calibration: bool, optional
    :ivar tth_initial_guess: Initial guess for 2&theta.
    :type tth_initial_guess: float, optional
    """
    calibration_method: Optional[Literal[
        'direct_fit_bragg', 'direct_fit_tth_ecc']] = 'direct_fit_bragg'
    quadratic_energy_calibration: Optional[bool] = False
    tth_initial_guess: Optional[
        confloat(gt=0, allow_inf_nan=False)] = Field(None, exclude=True)

    def flux_file_energy_range(self):
        """Get the energy range in the flux corection file.

        :return: The energy range in the flux corection file.
        :rtype: tuple(float, float)
        """
        if self.flux_file is None:
            return None
        flux = np.loadtxt(self.flux_file)
        energies = flux[:,0]/1.e3
        return energies.min(), energies.max()


class StrainAnalysisConfig(MCACalibrationConfig):
    """Class representing input parameters required to perform a
    strain analysis.

    :ivar find_peak_cutoff: Use scipy.signal.find_peaks to exclude
        peaks for all spectra for a given detector and user specified
        mask. A particular HKL peak is removed from the set of HKLs,
        when its mean peak height is  below `find_peak_cutoff` times
        the maximum mean intensity for that detector. Defaults to `0`
        in which case this step is ignored.
    :type find_peak_cutoff: float, optional
    :ivar oversampling: FIX
    :type oversampling: FIX
    :ivar rel_height_cutoff: Used to excluded peaks based on the
        `find_peak` parameter as well as for peak fitting exclusion
        of the individual detector spectra (see the strain detector
        configuration `CHAP.edd.models.MCADetectorStrainAnalysis).
        Defaults to `None`.
    :type rel_height_cutoff: float, optional
    :ivar skip_animation: Skip the animation and plotting of
        the strain analysis fits, defaults to `False`.
    :type skip_animation: bool, optional
    :ivar sum_axes: Whether to sum over the fly axis or not
        for EDD scan types not 0, defaults to `True`.
    :type sum_axes: Union[bool, list[str]], optional
    """
    find_peak_cutoff: Optional[confloat(ge=0.0, allow_inf_nan=False)] = 0.0
    num_proc: Optional[conint(gt=0)] = max(1, os.cpu_count()//4)
    oversampling: dict = {'num': 10}
    rel_height_cutoff: Optional[
        confloat(gt=0.0, lt=1.0, allow_inf_nan=False)] = None
    skip_animation: Optional[bool] = False
    sum_axes: Optional[
        Union[bool, conlist(min_length=1, item_type=str)]] = True

# FIX tth_file/tth_map not updated
#    @field_validator('detectors')
#    @classmethod
#    def validate_detectors(cls, detectors, info):
#        """Validate detector element tth_file field. It may only be
#        used if StrainAnalysisConfig used par_file.
#        """
#        for detector in detectors:
#            tth_file = detector.tth_file
#            if tth_file is not None:
#                if not info.data.get('par_file'):
#                    raise ValueError(
#                        'variable tth angles may only be used with a '
#                        'StrainAnalysisConfig that uses par_file.')
#                else:
#                    try:
#                        detector.tth_map = ParFile(
#                            info.data['par_file']).map_values(
#                                info.data['map_config'],
#                                np.loadtxt(tth_file))
#                    except Exception as e:
#                        raise ValueError(
#                            'Could not get map of tth angles from '
#                            f'{tth_file}') from e
#        return detectors

    @field_validator('oversampling')
    @classmethod
    def validate_oversampling(cls, oversampling, info):
        """Validate the oversampling field.

        :param oversampling: The value of `oversampling` to validate.
        :type oversampling: dict
        :param info: Pydantic validator info object.
        :type info: StrainAnalysisConfig,
            pydantic_core._pydantic_core.ValidationInfo
        :return: The validated value for oversampling.
        :rtype: bool
        """
        # Local modules
        from CHAP.utils.general import is_int

        raise ValueError('oversampling not updated yet')
        map_config = info.data.get('map_config')
        if map_config is None or map_config.attrs['scan_type'] < 3:
            return None
        if oversampling is None:
            return {'num': 10}
        if 'start' in oversampling and not is_int(oversampling['start'], ge=0):
            raise ValueError('Invalid "start" parameter in "oversampling" '
                             f'field ({oversampling["start"]})')
        if 'end' in oversampling and not is_int(oversampling['end'], gt=0):
            raise ValueError('Invalid "end" parameter in "oversampling" '
                             f'field ({oversampling["end"]})')
        if 'width' in oversampling and not is_int(oversampling['width'], gt=0):
            raise ValueError('Invalid "width" parameter in "oversampling" '
                             f'field ({oversampling["width"]})')
        if ('stride' in oversampling
                and not is_int(oversampling['stride'], gt=0)):
            raise ValueError('Invalid "stride" parameter in "oversampling" '
                             f'field ({oversampling["stride"]})')
        if 'num' in oversampling and not is_int(oversampling['num'], gt=0):
            raise ValueError('Invalid "num" parameter in "oversampling" '
                             f'field ({oversampling["num"]})')
        if 'mode' in oversampling and 'mode' not in ('valid', 'full'):
            raise ValueError('Invalid "mode" parameter in "oversampling" '
                             f'field ({oversampling["mode"]})')
        if not ('width' in oversampling or 'stride' in oversampling
                or 'num' in oversampling):
            raise ValueError('Invalid input parameters, specify at least one '
                             'of "width", "stride" or "num"')
        return oversampling



"""This subpackage contains generic utilities for fitting, parsing
CHESS scan data, collecting interactive user input, and finding
lattice properties of materials (among others).
"""

from CHAP.utils.fit import FitProcessor
